/**************************************************************************
     USBUA100 - USB driver module for the Roland UA-100, SC-8850 & UM-1

		(C) 2001 by Michael Minn - michael@michaelminn.com

   printk messages are written to /var/log/messages

   See /usr/src/linux/Documentation/usb for more information about USB programming

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

****************************************************************************/

#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/usb.h>
#include <linux/sound.h>
#include <linux/soundcard.h>

/******
*  To time isosynchronous writes correctly, URBs contain 10 packets totaling 441 samples
*  apiece (i.e. 10ms per URB). This works out to 3528 bytes for write URBs with 8 byte samples
*  (two 16-bit stereo channels)...nine packets of 352 bytes and one packet of 360 bytes
*
*  For Reads, however, each packet must be sent at full length (184 bytes) or you will get a nasty
*  kernel message in /var/log/messages: "usb-uhci.c: interrupt, status 2, frame# xxx"
*
******/

#define USBUA100_MAX_AUDIO_DEVICES 8
#define USBUA100_MAX_MIDI_DEVICES 8
#define USBUA100_AUDIO_READ_PACKET_LENGTH 184
#define USBUA100_AUDIO_WRITE_PACKET_LENGTH 352
#define USBUA100_AUDIO_PACKETS_PER_URB 10
#define USBUA100_AUDIO_URBS 16 /* smaller numbers result in occasional pauses due to delays reading from disk? */
#define USBUA100_AUDIO_READ_URB_BUFFER_LENGTH (USBUA100_AUDIO_READ_PACKET_LENGTH * USBUA100_AUDIO_PACKETS_PER_URB)
#define USBUA100_AUDIO_WRITE_URB_BUFFER_LENGTH ((USBUA100_AUDIO_WRITE_PACKET_LENGTH * USBUA100_AUDIO_PACKETS_PER_URB) + 8)

/************************ UA-100 Digital Audio ***********************************
*
*	Because the OSS API doesn't implement any way of storing private
*	data and only references devices by device number, context information
*	must be stored in this global array indexed by device number
*
***************************************************************************************/

struct usbua100_audio_device
{
	int channels;
	int is_ua100;
	int read_endpoint;
	int write_endpoint;
	int interface_number;
	int oss_minor_number;
	char *device_name;

	purb_t read_urb[USBUA100_AUDIO_URBS];
	int current_read_urb;
	int current_read_packet;
	int current_read_packet_offset;
	spinlock_t read_lock;

	purb_t write_urb[USBUA100_AUDIO_URBS];
	int current_write_urb;
	int current_write_packet;
	int current_write_packet_offset;
	int frame_for_next_write_urb;
	spinlock_t write_lock;

	wait_queue_head_t read_wait_queue;
	wait_queue_head_t write_wait_queue;

	struct usb_device *usb_device;
} usbua100_audio_devices[USBUA100_MAX_AUDIO_DEVICES];

struct file_operations usbua100_audio_operations;

static struct usbua100_audio_device *usbua100_assign_audio_device(struct usb_device *usb_device, char *device_name, int interface_number, int endpoint, int is_output)
{
	int scan = 0;
	struct usbua100_audio_device *device = NULL;

	for (scan = 0; (scan < USBUA100_MAX_AUDIO_DEVICES) && (!device); ++scan)
		if (usbua100_audio_devices[scan].is_ua100)
			device = &usbua100_audio_devices[scan];

	if ((!device) ||
		(device && is_output && device->write_endpoint) ||
		(device && (!is_output) && device->read_endpoint))
	{
		int oss_minor_number = register_sound_dsp(&usbua100_audio_operations, -1);
		device = &usbua100_audio_devices[oss_minor_number >> 4];
		memset(device, 0, sizeof(struct usbua100_audio_device));
		device->oss_minor_number = oss_minor_number;
	}

	if (!device)
	{
		printk(KERN_ERR "USBUA100: Failure registering audio device with OSS\n");
		return 0;
	}

	else if ((device->oss_minor_number >> 4) == 0)
		printk(KERN_INFO "USBA100: Registered %s as /dev/dsp - interface %d, endpoint %d\n", device_name, interface_number, endpoint);

	else
		printk(KERN_INFO "USBA100: Registered %s as /dev/dsp%d - interface %d, endpoint %d\n", device_name, device->oss_minor_number >> 4, interface_number, endpoint);

	device->is_ua100 = 1;
	device->usb_device = usb_device;
	device->interface_number = interface_number;
	device->device_name = device_name;

	if (is_output)
	{
		device->write_endpoint = endpoint;

		for (scan = 0; scan < USBUA100_AUDIO_URBS; ++scan)
		{
			int frame = 0;

			device->write_urb[scan] = usb_alloc_urb(USBUA100_AUDIO_PACKETS_PER_URB);
			device->write_urb[scan]->context = device;
			device->write_urb[scan]->transfer_buffer = kmalloc(USBUA100_AUDIO_WRITE_URB_BUFFER_LENGTH, GFP_KERNEL);
			device->write_urb[scan]->transfer_buffer_length = USBUA100_AUDIO_WRITE_URB_BUFFER_LENGTH;
			device->write_urb[scan]->actual_length = USBUA100_AUDIO_WRITE_URB_BUFFER_LENGTH;
			device->write_urb[scan]->number_of_packets = USBUA100_AUDIO_PACKETS_PER_URB;

			/* Each URB contains 3528 bytes (441 samples).  At 1ms per frame, URBs must be */
			/* timed to start every 10 frames (i.e. 10ms). */
			for (frame = 0; frame < USBUA100_AUDIO_PACKETS_PER_URB; ++frame)
			{
				device->write_urb[scan]->iso_frame_desc[frame].offset = frame * USBUA100_AUDIO_WRITE_PACKET_LENGTH;
				if (frame == 9)
					device->write_urb[scan]->iso_frame_desc[frame].length = USBUA100_AUDIO_WRITE_PACKET_LENGTH + 8;
				else
					device->write_urb[scan]->iso_frame_desc[frame].length = USBUA100_AUDIO_WRITE_PACKET_LENGTH;
				device->write_urb[scan]->iso_frame_desc[frame].actual_length = device->write_urb[scan]->iso_frame_desc[frame].length;
				device->write_urb[scan]->iso_frame_desc[frame].status = 0;
			}
		}
	}
	else
	{
		device->read_endpoint = endpoint;

		for (scan = 0; scan < USBUA100_AUDIO_URBS; ++scan)
		{
			int frame = 0;
			device->read_urb[scan] = usb_alloc_urb(USBUA100_AUDIO_PACKETS_PER_URB);
			device->read_urb[scan]->context = device;
			device->read_urb[scan]->transfer_buffer = kmalloc(USBUA100_AUDIO_READ_URB_BUFFER_LENGTH, GFP_KERNEL);
			device->read_urb[scan]->transfer_buffer_length = USBUA100_AUDIO_READ_URB_BUFFER_LENGTH;
			device->read_urb[scan]->actual_length = device->read_urb[scan]->transfer_buffer_length;
			device->read_urb[scan]->number_of_packets = USBUA100_AUDIO_PACKETS_PER_URB;

			for (frame = 0; frame < USBUA100_AUDIO_PACKETS_PER_URB; ++frame)
			{
				device->read_urb[scan]->iso_frame_desc[frame].offset = frame * USBUA100_AUDIO_READ_PACKET_LENGTH;
				device->read_urb[scan]->iso_frame_desc[frame].length = USBUA100_AUDIO_READ_PACKET_LENGTH;
				device->read_urb[scan]->iso_frame_desc[frame].actual_length = device->read_urb[scan]->iso_frame_desc[frame].length;
				device->read_urb[scan]->iso_frame_desc[frame].status = 0;
			}
		}
	}

	return device;
}

static int usbua100_release_audio_device(struct usbua100_audio_device *device)
{
	int scan = 0;

	printk(KERN_INFO "USBUA100: Disconnecting AUDIO device 0x%02x\n", device - usbua100_audio_devices);

	device->usb_device = NULL;
	device->channels = 0;
	unregister_sound_dsp(device->oss_minor_number);

	for (scan = 0; scan < USBUA100_AUDIO_URBS; ++scan)
	{
		if (device->read_urb[scan])
		{
			usb_unlink_urb(device->read_urb[scan]);
			kfree(device->read_urb[scan]->transfer_buffer);
			usb_free_urb(device->read_urb[scan]);
		}
		if (device->write_urb[scan])
		{
			usb_unlink_urb(device->write_urb[scan]);
			kfree(device->write_urb[scan]->transfer_buffer);
			usb_free_urb(device->write_urb[scan]);
		}
	}

	memset(device, 0, sizeof(*device));

	return 0;
}

static int usbua100_audio_open (struct inode *inode, struct file *file)
{
	int blank_urb = 0;
	int blank_packet = 0;
	int status = 0;
	struct usbua100_audio_device *device = NULL;

	printk(KERN_INFO "USBUA100: UA-100 Wave Open\n");

	if ((MINOR(inode->i_rdev) >> 4) >= USBUA100_MAX_AUDIO_DEVICES)
		return -EINVAL;

	device = &usbua100_audio_devices[MINOR(inode->i_rdev) >> 4];
	file->private_data = device;

	if (!device->usb_device)
		return -EHOSTDOWN;

	if (device->channels)
		return -EBUSY;

	if ((status = usb_set_interface(device->usb_device, 0, 1)) < 0)
	{
		printk(KERN_ERR "USBUA100: Failure %d setting UA-100 incoming interface\n", status);
		return status;
	}

	if ((status = usb_set_interface(device->usb_device, 1, 1)) < 0)
	{
		printk(KERN_ERR "USBUA100: Failure %d setting UA-100 outgoing interface\n", status);
		return status;
	}

	init_waitqueue_head(&device->read_wait_queue);
	init_waitqueue_head(&device->write_wait_queue);

	device->channels = 2;
	device->frame_for_next_write_urb = 0;
	device->current_read_urb = 0;
	device->current_read_packet = 0;
	device->current_read_packet_offset = 0;
	device->current_write_urb = 0;
	device->current_write_packet = 0;
	device->current_write_packet_offset = 0;

	/* set packets initially empty so the first read immediately queues all buffers */
	for (blank_urb = 0; blank_urb < USBUA100_AUDIO_URBS; ++blank_urb)
	{
		memset(device->read_urb[blank_urb]->transfer_buffer, 0, USBUA100_AUDIO_READ_URB_BUFFER_LENGTH);
		for (blank_packet = 0; blank_packet < USBUA100_AUDIO_PACKETS_PER_URB; ++blank_packet)
			device->read_urb[blank_urb]->iso_frame_desc[blank_packet].actual_length = 0;
	}

	MOD_INC_USE_COUNT;

	return 0;
}

static int usbua100_audio_release (struct inode *inode, struct file *file)
{
	int scan;
	struct usbua100_audio_device *device = file->private_data;

	printk(KERN_INFO "USBUA100: UA-100 Wave Close\n");

	if (device)
		device->channels = 0;

	for (scan = 0; device && device->usb_device && (scan < USBUA100_AUDIO_URBS); ++scan)
	{
		if (device->read_urb && device->read_urb[scan])
			usb_unlink_urb(device->read_urb[scan]);
		if (device->write_urb && device->write_urb[scan])
			usb_unlink_urb(device->write_urb[scan]);
	}

    MOD_DEC_USE_COUNT;

	return 0;
}

static void usbua100_audio_write_callback(urb_t *urb)
{
	struct usbua100_audio_device *device = (struct usbua100_audio_device *) urb->context;

	if (waitqueue_active(&device->write_wait_queue))
		wake_up_interruptible(&device->write_wait_queue);
}

static ssize_t usbua100_audio_write (struct file *file, const char *data, size_t size, loff_t *offset)
{
	unsigned long flags;
	int user_data_transferred = 0;
	struct usbua100_audio_device *device = file->private_data;

	if ((!device) || (!device->usb_device))
		return -EHOSTDOWN;

	else if (size < (device->channels * 2))
		return -EINVAL;

	spin_lock_irqsave(&device->write_lock, flags);

	while (user_data_transferred < size)
	{
		int retry = 40;
		purb_t current_urb = device->write_urb[device->current_write_urb];

		/* Changing windows in KDE (page fault?) was causing this to wake up before URB */
		/* returned causing a false timeout condition.  This causes occasional playback */
		/* pauses, but at least writes don't fail and stop playback */
		while ((current_urb->status == -EINPROGRESS) && retry && device->usb_device)
		{
			interruptible_sleep_on_timeout(&device->write_wait_queue, HZ / 20);
			--retry;
		}

		if (!device->usb_device)
		{
			printk(KERN_ERR "Device disconnected during audio write wait\n");
			spin_unlock_irqrestore(&device->write_lock, flags);
			return -EHOSTDOWN;
		}

		if (current_urb->status == -EINPROGRESS)
		{
			printk(KERN_ERR "Timed out writing %d of %d bytes\n", user_data_transferred, size);
			spin_unlock_irqrestore(&device->write_lock, flags);
			return -ETIMEDOUT;
		}

		/* UA-100 interleaves 16-bit/44Khz samples for the two output devices: L1, R1, L2, R2*/
		/* If input is stereo, USBUA100 sends the data only to output device WAVE 1 */
		/* If input is mono, USBUA100 sends the data to both channels of WAVE 1 */
		if (device->channels == 4)
		{
			copy_from_user(&((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_write_packet].offset + device->current_write_packet_offset], &data[user_data_transferred], 8);
			user_data_transferred += 8;
			device->current_write_packet_offset += 8;
		}
		else if (device->channels == 2)
		{
			copy_from_user(&((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_write_packet].offset + device->current_write_packet_offset], &data[user_data_transferred], 4);
			user_data_transferred += 4;
			device->current_write_packet_offset += 4;
			memset(&((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_write_packet].offset + device->current_write_packet_offset], 0, 4);
			device->current_write_packet_offset += 4;
		}
		else
		{
			copy_from_user(&((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_write_packet].offset + device->current_write_packet_offset], &data[user_data_transferred], 2);
			copy_from_user(&((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_write_packet].offset + device->current_write_packet_offset + 2], &data[user_data_transferred], 2);
			user_data_transferred += 2;
			device->current_write_packet_offset += 4;
			memset(&((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_write_packet].offset + device->current_write_packet_offset], 0, 4);
			device->current_write_packet_offset += 4;
		}

		if (device->current_write_packet_offset >= current_urb->iso_frame_desc[device->current_write_packet].actual_length)
		{
			device->current_write_packet_offset = 0;
			++device->current_write_packet;
		}

		if (device->current_write_packet >= USBUA100_AUDIO_PACKETS_PER_URB)
		{
			device->current_write_packet = 0;
			++device->current_write_urb;
		}

		if (device->current_write_urb >= USBUA100_AUDIO_URBS)
			device->current_write_urb = 0;

		if ((device->current_write_packet == 0) && (device->current_write_packet_offset == 0))
		{
			int status = 0;
			int retry = 10;

			spin_lock_init(current_urb->lock);
			current_urb->status = 0;
			current_urb->context = device;
			current_urb->dev = device->usb_device;
			current_urb->transfer_flags = 0;
			current_urb->complete = usbua100_audio_write_callback;
			current_urb->pipe = usb_sndisocpipe(device->usb_device, device->write_endpoint);

			/* usb_submit_urb() will sporadically */
			/* return -EAGAIN...because start frame is too early?  Perhaps because */
			/* of the scheduler delay between when start_frame is set and submit is called? */
			do
			{
				if (device->frame_for_next_write_urb < (usb_get_current_frame_number (device->usb_device) + 10))
					device->frame_for_next_write_urb = usb_get_current_frame_number (device->usb_device) + 20;
				current_urb->start_frame = device->frame_for_next_write_urb;
				device->frame_for_next_write_urb += 10;
				--retry;
			}
			while (((status = usb_submit_urb(current_urb)) == -EAGAIN) && retry);

			if (status != 0)
				printk(KERN_ERR "USBUA100: failure %d submitting write_urb\n", status);
		}
	}

	spin_unlock_irqrestore(&device->write_lock, flags);

	return size;
}

static void usbua100_audio_read_callback(urb_t *urb)
{
	struct usbua100_audio_device *device = (struct usbua100_audio_device *) urb->context;

	/* printk(KERN_INFO "USBUA100: Read URB callback status %d\n", urb->status);*/
	if (waitqueue_active(&device->read_wait_queue))
		wake_up_interruptible(&device->read_wait_queue);
}

static ssize_t usbua100_audio_read (struct file *file, char *data, size_t size, loff_t *offset)
{
	unsigned long flags;
	int user_data_transferred = 0;
	struct usbua100_audio_device *device = file->private_data;

	if ((!device) || (!device->usb_device))
		return -EHOSTDOWN;

	else if (size < 4)
		return -EINVAL;

	spin_lock_irqsave(&device->read_lock, flags);

	while (user_data_transferred < size)
	{
		purb_t current_urb = device->read_urb[device->current_read_urb];

		if (current_urb->status == -EINPROGRESS)
			interruptible_sleep_on_timeout(&device->read_wait_queue, HZ * 5);

		if (!device->usb_device)
		{
			printk(KERN_ERR "Device disconnected during audio read wait\n");
			spin_unlock_irqrestore(&device->read_lock, flags);
			return -EHOSTDOWN;
		}

		if (current_urb->status == -EINPROGRESS)
		{
			printk(KERN_ERR "Timed out reading %d of %d bytes\n", user_data_transferred, size);
			spin_unlock_irqrestore(&device->read_lock, flags);
			return -ETIMEDOUT;
		}

		copy_to_user(&data[user_data_transferred], &((char*)current_urb->transfer_buffer)[current_urb->iso_frame_desc[device->current_read_packet].offset + device->current_read_packet_offset], 4);
		device->current_read_packet_offset += 4;
		user_data_transferred += 4;

		if (device->current_read_packet_offset >= current_urb->iso_frame_desc[device->current_read_packet].actual_length)
		{
			device->current_read_packet_offset = 0;
			++device->current_read_packet;
		}

		if (device->current_read_packet >= USBUA100_AUDIO_PACKETS_PER_URB)
		{
			device->current_read_packet = 0;
			++device->current_read_urb;
		}

		if (device->current_read_urb >= USBUA100_AUDIO_URBS)
			device->current_read_urb = 0;

		/* When done reading the data from a URB - resubmit it */
		if ((device->current_read_packet == 0) && (device->current_read_packet_offset == 0))
		{
			int status = 0;

			spin_lock_init(current_urb->lock);
			current_urb->status = 0;
			current_urb->start_frame = 0;
			current_urb->context = device;
			current_urb->dev = device->usb_device;
			current_urb->transfer_flags = USB_ISO_ASAP;
			current_urb->complete = usbua100_audio_read_callback;
			current_urb->pipe = usb_rcvisocpipe(device->usb_device, device->read_endpoint);

			if ((status = usb_submit_urb(current_urb)) != 0)
			{
				printk(KERN_ERR "USBUA100: failure %d submitting read urb\n", status);
				spin_unlock_irqrestore(&device->read_lock, flags);
				return status;
			}
		}
	}

	spin_unlock_irqrestore(&device->read_lock, flags);
	return size;
}

static int usbua100_audio_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	int data_value = 0;
	struct usbua100_audio_device *device = file->private_data;

	if (data)
		data_value = *((int*)data);

	if ((command == SOUND_PCM_WRITE_BITS) && (data_value == 16))
		return 0;

	else if ((command == SOUND_PCM_WRITE_CHANNELS) && (data_value == 1))
	{
		device->channels = 1;
		return 0;
	}

	else if ((command == SOUND_PCM_WRITE_CHANNELS) && (data_value == 2))
	{
		device->channels = 2;
		return 0;
	}

	else if ((command == SOUND_PCM_WRITE_CHANNELS) && (data_value == 4))
	{
		device->channels = 4;
		return 0;
	}

	else if ((command == SOUND_PCM_WRITE_RATE) && (data_value == 44100))
		return 0;

	else if (command == SOUND_PCM_READ_RATE)
		return (data = 44100);

	else if (command == SOUND_PCM_READ_CHANNELS)
		return (data = 2);

	else if (command == SOUND_PCM_READ_BITS)
		return (data = 16);

	else if (command == SNDCTL_DSP_RESET)
		return 0; /* don't do anything, but acknowledge so sox will work */

	else if (command == SNDCTL_DSP_SYNC)
		return 0; /* don't do anything, but acknowledge so sox will work */

	else if (command == SNDCTL_DSP_STEREO)
		return 0; /* don't do anything, but acknowledge so standard audio applications will work */

	else if (command == SNDCTL_DSP_SETFRAGMENT)
		return 0; /* don't do anything, but acknowledge so standard audio applications will work */

	else if (command == SNDCTL_DSP_GETBLKSIZE)
	{
		*((int*)data) = USBUA100_AUDIO_WRITE_URB_BUFFER_LENGTH * 2;
		return 0;
	}

	else if (command == SNDCTL_MIDI_INFO)
	{
		struct midi_info midi_info;
		memset(&midi_info, 0, sizeof(struct midi_info));
		strcpy(midi_info.name, device->device_name);
		copy_to_user(((struct midi_info*) data), &midi_info, sizeof(struct midi_info));
		return 0;
	}

	else
	{
		printk(KERN_INFO "USBUA100: Undefined audio IOCTL %x\n", command);
		return -1;
	}

}

struct file_operations usbua100_audio_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	usbua100_audio_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	usbua100_audio_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	usbua100_audio_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	usbua100_audio_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	usbua100_audio_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};


/************************* MIDI Event Functions *****************************/

#define USBUA100_MIDI_NOTE_OFF 0x80
#define USBUA100_MIDI_NOTE_ON  0x90
#define USBUA100_MIDI_POLYPHONIC_KEY_PRESSURE 0xA0
#define USBUA100_MIDI_CONTROL_CHANGE 0xB0
#define USBUA100_MIDI_PROGRAM_CHANGE 0xC0
#define USBUA100_MIDI_AFTERTOUCH 0xD0
#define USBUA100_MIDI_PITCH_BEND 0xE0
#define USBUA100_MIDI_SYSTEM_EXCLUSIVE 0xF0
#define USBUA100_MIDI_TIME_CODE_QUARTER_FRAME 0xF1
#define USBUA100_MIDI_SONG_POSITION 0xF2
#define USBUA100_MIDI_SONG_SELECT 0xF3
#define USBUA100_MIDI_TUNE_REQUEST 0xF6
#define USBUA100_MIDI_END_OF_EXCLUSIVE 0xF7
#define USBUA100_MIDI_TIMING_CLOCK 0xF8
#define USBUA100_MIDI_START 0xFA
#define USBUA100_MIDI_CONTINUE 0xFB
#define USBUA100_MIDI_STOP 0xFC
#define USBUA100_MIDI_ACTIVE_SENSING 0xFE
#define USBUA100_MIDI_META_EVENT 0xFF

static int usbua100_midi_message_length(unsigned char *message, int message_length)
{
	if (!(message[0] & 0x80))
		return message_length; /* if the message does not begin with a status byte, assume it's incomplete and needs to be sent immediately */

	else switch (message[0] & 0xF0)
	{
		case USBUA100_MIDI_NOTE_OFF:
		case USBUA100_MIDI_NOTE_ON:
		case USBUA100_MIDI_POLYPHONIC_KEY_PRESSURE:
		case USBUA100_MIDI_CONTROL_CHANGE:
		case USBUA100_MIDI_PITCH_BEND:
			return 3;

		case USBUA100_MIDI_PROGRAM_CHANGE:
		case USBUA100_MIDI_AFTERTOUCH:
			return 2;

		default:
			switch (message[0])
			{
				case USBUA100_MIDI_SONG_POSITION:
					return 3;

				case USBUA100_MIDI_TIME_CODE_QUARTER_FRAME:
				case USBUA100_MIDI_SONG_SELECT:
					return 2;

				case USBUA100_MIDI_TUNE_REQUEST:
				case USBUA100_MIDI_END_OF_EXCLUSIVE:
				case USBUA100_MIDI_TIMING_CLOCK:
				case USBUA100_MIDI_START:
				case USBUA100_MIDI_CONTINUE:
				case USBUA100_MIDI_STOP:
				case USBUA100_MIDI_ACTIVE_SENSING:
					return 1;

				case USBUA100_MIDI_META_EVENT:
					if (message_length <= 2)
						return 3; /* a meta event is at least three bytes long */
					else
						return message[2] + 3;

				case USBUA100_MIDI_SYSTEM_EXCLUSIVE:
					if (message_length <= 0)
						return 2; /* Sys Ex is at least two bytes long */
					else if (message[message_length - 1] == USBUA100_MIDI_END_OF_EXCLUSIVE)
						return message_length;
					else
						return message_length + 1;
			}
	}

	return message_length;
}


/************************* USB Debug Functions *****************************/

/***
static void usbua100_urb_dump(char *message, urb_t *urb)
{
	unsigned char *char_data = (unsigned char*) urb->transfer_buffer;
	int scan = 0;

	printk(KERN_INFO "USBUA100: %s\n", message);
	if (!urb)
	{
		printk(KERN_INFO "NULL URB\n");
		return;
	}

	printk(KERN_INFO " pipe %x, transfer_flags %x", urb->pipe, urb->transfer_flags);
	printk(" status %d, length %d, actual %d\n", urb->status, urb->transfer_buffer_length, urb->actual_length);

	for (scan = 0; scan < urb->transfer_buffer_length; ++scan)
	{
		printk (" %02x", (int) char_data[scan]);
		if ((scan & 0xf) == 0xf)
			printk("\n");
  	}

	printk("\n");
}
***/

/****
static void usbua100_device_dump(char *message, struct usb_device *device)
{
	printk(KERN_INFO "USBUA100: %s\n", message);
	printk(KERN_INFO "  Device #%d, slow %d\n", device->devnum, device->slow);
	printk(KERN_INFO "  Alternate #%d, endpoints %d\n",
		device->actconfig->interface[2].act_altsetting,
		device->actconfig->interface[2].altsetting[device->actconfig->interface->act_altsetting].bNumEndpoints);
}
*/

/************************ OSS Registration ***********************************
*
*	Because the OSS API doesn't implement any way of storing private
*	data and only references devices by device number, context information
*	must be stored in this global array indexed by device number
*
***************************************************************************************/

struct usbua100_midi_device
{
	struct usb_device *usb_device;
	unsigned char burst_byte;
	int is_open;
	int read_endpoint_address;
	int write_endpoint_address;
	int urb_buffer_length;
	purb_t read_urb;
	purb_t write_urb;
	wait_queue_head_t read_wait_queue;
	wait_queue_head_t write_wait_queue;
	int oss_minor_number;
	char *device_name;
} usbua100_midi_devices[USBUA100_MAX_MIDI_DEVICES];

struct file_operations usbua100_midi_file_operations;

static struct usbua100_midi_device *usbua100_assign_midi_device(struct usb_device *usb_device, char *device_name, int in_endpoint, int out_endpoint, unsigned char burst_byte, int buffer_length)
{
	int status = 0;
	int oss_minor_number = register_sound_midi(&usbua100_midi_file_operations, -1);
	struct usbua100_midi_device *device = NULL;

	if (oss_minor_number < 0)
	{
		printk(KERN_ERR "NOTEMIDI: Failure %d registering with OSS\n", oss_minor_number);
		return 0;
	}
	else
		printk(KERN_INFO "USBA100: Registered %s as /dev/midi%02d\n", device_name, oss_minor_number >> 4);

	device = &usbua100_midi_devices[oss_minor_number >> 4];
	memset(device, 0, sizeof(struct usbua100_midi_device));
	device->oss_minor_number = oss_minor_number;
	device->device_name = device_name;

	device->usb_device = usb_device;
	device->read_endpoint_address = in_endpoint;
	device->write_endpoint_address = out_endpoint;
	device->burst_byte = burst_byte;
	device->urb_buffer_length = buffer_length;
	device->is_open = 0;

	/* The buffer sent with a URB must be allocated with kmalloc */
	/* or you get an interrupt error from usb-uhci - page alignment perhaps? */
	/* For reads, the buffer must be the EXACT length specified by the */
	/* device descriptor */
	device->read_urb = usb_alloc_urb(0);
	device->read_urb->transfer_buffer = (unsigned char *)kmalloc(buffer_length, GFP_KERNEL);
	device->read_urb->context = device;
	device->read_urb->transfer_buffer_length = buffer_length;

	device->write_urb = usb_alloc_urb(0);
	device->write_urb->transfer_buffer = (unsigned char *)kmalloc(buffer_length, GFP_KERNEL);
	device->write_urb->context = device;
	device->write_urb->transfer_buffer_length = buffer_length;

	if ((status = usb_set_interface(usb_device, 2, 0)) < 0)
	{
		printk(KERN_ERR "USBUA100: Failure %d setting interface\n", status);
		return NULL;
	}

	return device;
}

static int usbua100_release_midi_device(struct usbua100_midi_device *device)
{
	printk(KERN_INFO "USBUA100: Disconnecting MIDI device 0x%02x\n", device - usbua100_midi_devices);

	device->usb_device = NULL;
	device->is_open = 0;
	unregister_sound_midi(device->oss_minor_number);

	if (device->write_urb)
	{
		usb_unlink_urb(device->write_urb);
		kfree(device->write_urb->transfer_buffer);
		usb_free_urb(device->write_urb);
	}

	if (device->read_urb)
	{
		usb_unlink_urb(device->read_urb);
		kfree(device->read_urb->transfer_buffer);
		usb_free_urb(device->read_urb);
	}

	return 0;
}

/************************ Roland MIDI Protocol ***********************************
*
*	MIDI data is sent and received in 4-byte chunks with a preceeding "burst"
*	byte that indicates the destination/source port for following 3 bytes of
*	data. All MIDI messages under status 0xf0 (SYSEX) are two or three bytes long
*	and fit in a single 4-byte chunk. Messages over status byte 0xf0 are split into
*	multiple consecutive chunks with the appropriate burst bytes specified below.
*
*	For MIDI sends, the upper 4 bits of the "burst" byte are the source/destination
*	port (described below). The lower 4 bits are either the upper 4 bits of the MIDI
*	message status byte, 4 for the beginning or middle of a SYSEX message, 6 if there
*	are two bytes in the last SYSEX chunk, and 5 if there is one byte in the last
*	chunk of a SYSEX message. For MIDI receives the protocol is basically
*	the same with the upper four bits representing the input port.
*
*		Example: 19 90 30 70 is a note on message sent to port 2.
*
*	For UA-100 sends, the "burst byte" is 0x00 for MIDI OUT 1, 0x10 for MIDI OUT 2
*	and 0x20 for the UA-100 control port which receives SYSEX messages to change
*	the internal mixer and effect settings. MIDI receives are basically the same:
*	0x00 for MIDI IN 1 and 0x00 for MIDI IN 2.
*
*	For the UM-1, there is only one input and output port, so the upper 4 bits of
*	the "burst byte" are always 0x00.
*
*	For the SC-8850, the "burst byte" indicates which of the four "parts" (i.e.
*	16-channel banks) should receive the MIDI output: 0x00 for Part A, 0x10
*	for Part B, 0x20 for Part C and 0x30 for Part D.  However, unlike the other
*	devices, for MIDI input: 0x40 for MIDI IN 1 and 0x50 for MIDI IN 2
*
*	Chunks smaller than 4 bytes (such as two-byte program change messages) are
*	padded with trailing NULLs to a 4-byte length.
*
***************************************************************************************/

static int usbua100_midi_file_open (struct inode *inode, struct file *file)
{
	int status = 0;
	struct usbua100_midi_device *device = NULL;

	if ((MINOR(inode->i_rdev) >> 4) >= USBUA100_MAX_MIDI_DEVICES)
		return -EINVAL;

	file->private_data = (void*)(MINOR(inode->i_rdev) >> 4);
	device = &usbua100_midi_devices[MINOR(inode->i_rdev) >> 4];

	if (!device->usb_device)
		return -EHOSTDOWN; /* device unplugged? */

	printk(KERN_INFO "USBUA100: %s opened /dev/midi%02d\n", device->device_name, device - usbua100_midi_devices);

	init_waitqueue_head(&device->read_wait_queue);
	init_waitqueue_head(&device->write_wait_queue);

	/***
	if ((halt_status = usb_clear_halt(device->usb_device, usb_rcvbulkpipe(device->usb_device, device->read_endpoint_address))) < 0)
		printk(KERN_INFO "USBUA100: Failure %d clearing input halt\n", halt_status);
	***/

	device->read_urb->status = 0;
	device->write_urb->status = 0;
	++device->is_open;

	MOD_INC_USE_COUNT;

	return status;
}

static int usbua100_midi_file_release (struct inode *inode, struct file *file)
{
	struct usbua100_midi_device *device = &usbua100_midi_devices[MINOR(inode->i_rdev) >> 4];

	printk(KERN_INFO "USBUA100: %s closed /dev/midi%02d\n", device->device_name, device - usbua100_midi_devices);

	if (device && device->usb_device && device->is_open)
	{
		usb_unlink_urb(device->read_urb);
		usb_unlink_urb(device->write_urb);
		--device->is_open;
	}

    MOD_DEC_USE_COUNT;

	return 0;
}

static void usbua100_midi_write_callback(urb_t *urb)
{
	struct usbua100_midi_device *device = (struct usbua100_midi_device *) urb->context;

	if (waitqueue_active(&device->write_wait_queue))
		wake_up_interruptible(&device->write_wait_queue);
}

static ssize_t usbua100_midi_file_write (struct file *file, const char *user_data, size_t length, loff_t *offset)
{
	int status = 0;
	int message_length = 0;
	int block_start = 0;
	unsigned char data[256];
	struct usbua100_midi_device *device = &usbua100_midi_devices[(int)file->private_data];

	if (length > 256)
		length = 256;

	copy_from_user(data, user_data, length);
	message_length = usbua100_midi_message_length(data, length);

	/* printk(KERN_INFO "USBUA100: Write %02x to device %02x, buffer length %d ------------------- \n", data, device - usbua100_midi_devices, device->write_buffer_length); */

	if ((!device) || (!device->usb_device) || (!device->is_open))
		return -EHOSTDOWN; /* device unplugged? */

	else if ((data[0] < 0x80) || (data[0] > 0xf4))
		return -EINVAL; /* ignore MIDI messages without status byte or status greater than SYSEX or Roland part select F4 */

	else if (length < message_length)
		return -EINVAL; /* Ignore incomplete MIDI messages */

	/* The UA-100 limits bulk transfers to 32 bytes per URB (i.e. 24 bytes of data + the burst bytes) */
	/* Therefore, long SYSEX messages (such as the 40 byte effect parameter messages) must be broken */
	/* into multiple successive URBs */
	for (block_start = 0; block_start < length; block_start += (device->urb_buffer_length * 3 / 4))
	{
		int chunk_start = 0;
		int block_end = block_start + (device->urb_buffer_length * 3 / 4);

		if (block_end > length)
			block_end = length;

		if (device->write_urb->status == -EINPROGRESS)
			interruptible_sleep_on_timeout(&device->write_wait_queue, HZ);

		if (device->write_urb->status == -EINPROGRESS)
			return -ETIMEDOUT;

		device->write_urb->transfer_buffer_length = 0;
		for (chunk_start = block_start; chunk_start < block_end; chunk_start += 3)
		{
			int chunk_length = 3;
			if (chunk_length > (length - chunk_start))
				chunk_length = length - chunk_start;

			if (data[0] != 0xf0)
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = ((data[chunk_start] >> 4) & 0x0f) | device->burst_byte;
			else if ((chunk_start + 3) < length)
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 0x04 | device->burst_byte;
			else if ((length - chunk_start) == 1)
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 0x05 | device->burst_byte;
			else
				((unsigned char*)device->write_urb->transfer_buffer) [device->write_urb->transfer_buffer_length] = 0x06 | device->burst_byte;

			memset(&((unsigned char*)device->write_urb->transfer_buffer)[device->write_urb->transfer_buffer_length + 1], 0, 3); /* chunk length must be multiple of 4? */
			memcpy(&((unsigned char*)device->write_urb->transfer_buffer)[device->write_urb->transfer_buffer_length + 1], &data[chunk_start], chunk_length);
			device->write_urb->transfer_buffer_length += 4;
		}

		spin_lock_init(device->write_urb->lock);
		device->write_urb->dev = device->usb_device;
		device->write_urb->pipe = usb_sndbulkpipe(device->usb_device, device->write_endpoint_address);
		device->write_urb->complete = (usb_complete_t)usbua100_midi_write_callback;
		device->write_urb->context = device;

		/*usbua100_urb_dump("MIDI write", device->write_urb);*/

		if ((status = usb_submit_urb(device->write_urb)) != 0)
		{
			printk(KERN_ERR "USBUA100: failure %d submitting write_urb (%d bytes)\n", status, device->write_urb->transfer_buffer_length);
			return status;
		}
		else
			status = length;
	}

	return status;
}

static void usbua100_midi_read_callback(urb_t *urb)
{
	struct usbua100_midi_device *device = (struct usbua100_midi_device *) urb->context;

	if (waitqueue_active(&device->read_wait_queue))
		wake_up_interruptible(&device->read_wait_queue);
}

static ssize_t usbua100_midi_file_read (struct file *file, char *data, size_t length, loff_t *offset)
{
	int scan = 0;
	int status = 0;
	int actual_length = 0;
	struct usbua100_midi_device *device = &usbua100_midi_devices[(int)file->private_data];

	if ((!device) || (!device->usb_device) || (!device->is_open))
		return -EHOSTDOWN; /* device unplugged? */

	spin_lock_init(device->read_urb->lock);
	device->read_urb->dev = device->usb_device;
	device->read_urb->pipe = usb_rcvbulkpipe(device->usb_device, device->read_endpoint_address);
	((unsigned char*)device->read_urb->transfer_buffer)[0] = 0;
	device->read_urb->complete = (usb_complete_t)usbua100_midi_read_callback;
	device->read_urb->context = device;

	if ((status = usb_submit_urb(device->read_urb)) < 0)
	{
		printk(KERN_ERR "USBUA100: failure %d submitting read_urb to endpoint %02x\n", status, device->read_endpoint_address);
		return status;
	}

	interruptible_sleep_on(&device->read_wait_queue);

	/* usbua100_urb_dump("Read URB -----------", device->read_urb); */

	if ((!device) || (!device->usb_device) || (!device->is_open))
		return -EHOSTDOWN; /* device unplugged? */

	else if (((unsigned char*)device->read_urb->transfer_buffer)[0] == 0)
		actual_length = 0; /* UA-100 regularly returns empty buffers */

	/*
	else if ((((unsigned char*)device->read_urb->transfer_buffer)[0] & 0xf0) != device->burst_byte)
		length = 0; data from a different port - doesn't work with SC-8850
	*/

	else for (scan = 0; (scan < device->read_urb->actual_length) && (((scan / 4) * 3) < length) && (((unsigned char*)device->read_urb->transfer_buffer) [scan] > 0); scan += 4)
		switch (((unsigned char*)device->read_urb->transfer_buffer) [scan] & 0x0f)
		{
			case 0x05: /* 1 byte end of sysex */
			case 0x0F: /* single byte ? */
				copy_to_user(&data[(scan / 4) * 3], &((unsigned char*)device->read_urb->transfer_buffer) [scan + 1], 1);
				actual_length += 1;
				break;

			case 0x02: /* 2 byte system common message */
			case 0x06: /* 2 byte end of sysex */
			case 0x0C: /* Program Change*/
			case 0x0D: /* Aftertouch */
				copy_to_user(&data[(scan / 4) * 3], &((unsigned char*)device->read_urb->transfer_buffer) [scan + 1], 2);
				actual_length += 2;
				break;

			default: /* everything else is 3-byte packet */
				copy_to_user(&data[(scan / 4) * 3], &((unsigned char*)device->read_urb->transfer_buffer) [scan + 1], 3);
				actual_length += 3;
				break;
		}

	return actual_length;
}

static int usbua100_midi_file_ioctl (struct inode *inode, struct file *file, unsigned int command, unsigned long data)
{
	struct usbua100_midi_device *device = &usbua100_midi_devices[(int)file->private_data];
	struct midi_info midi_info;

	if (command == SNDCTL_MIDI_INFO)
	{
		memset(&midi_info, 0, sizeof(struct midi_info));
		strcpy(midi_info.name, device->device_name);
		copy_to_user(((struct midi_info*) data), &midi_info, sizeof(struct midi_info));
		return 0;
	}

	else
	{
		printk(KERN_INFO "USBUA100: Undefined MIDI IOCTL %x\n", command);
		return -1;
	}
}




struct file_operations usbua100_midi_file_operations =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* struct module *owner; */
#endif
	NULL, /* loff_t (*llseek) (struct file *, loff_t, int); */
	usbua100_midi_file_read, /* ssize_t (*read) (struct file *, char *, size_t, loff_t *); */
	usbua100_midi_file_write, /* ssize_t (*write) (struct file *, const char *, size_t, loff_t *); */
	NULL, /* int (*readdir) (struct file *, void *, filldir_t); */
	NULL, /* unsigned int (*poll) (struct file *, struct poll_table_struct *); */
	usbua100_midi_file_ioctl, /* int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long); */
	NULL, /* int (*mmap) (struct file *, struct vm_area_struct *); */
	usbua100_midi_file_open, /* int (*open) (struct inode *, struct file *); */
	NULL, /* int (*flush) (struct file *); */
	usbua100_midi_file_release, /* int (*release) (struct inode *, struct file *); */
	NULL, /* int (*fsync) (struct file *, struct dentry *); */
	NULL, /* int (*fasync) (int, struct file *, int); */
	NULL, /* int (*lock) (struct file *, int, struct file_lock *); */
	NULL, /* ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *); */
	NULL, /* ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *); */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	NULL, /* ssize_t (*writepage) (struct file *, struct page *, int, size_t, loff_t *, int); */
#endif
};



/************************* USB Driver Registration *****************************/

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
static void *usbua100_probe(struct usb_device *device, unsigned interface, const struct usb_device_id *id)
#else
static void *usbua100_probe(struct usb_device *device, unsigned interface)
#endif
{
	if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 0) && (interface == 0))
		return usbua100_assign_audio_device(device, "UA-100 Out", interface, 0x01, 1);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 0) && (interface == 1))
		return usbua100_assign_audio_device(device, "UA-100 In", interface, 0x81, 0);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 0) && (interface == 2))
	{
		usbua100_assign_midi_device(device, "UA-100 A", 0x82, 0x02, 0x00, 32);
		usbua100_assign_midi_device(device, "UA-100 B", 0x82, 0x02, 0x10, 32);
		return usbua100_assign_midi_device(device, "UA-100 CTRL", 0x82, 0x02, 0x20, 32);
	}

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 3) && (interface == 2))
	{
		usbua100_assign_midi_device(device, "SC-8850 A", 0x81, 0x01, 0x00, 64);
		usbua100_assign_midi_device(device, "SC-8850 B", 0x81, 0x01, 0x10, 64);
		usbua100_assign_midi_device(device, "SC-8850 C", 0x81, 0x01, 0x20, 64);
		return usbua100_assign_midi_device(device, "SC-8850 D", 0x81, 0x01, 0x30, 64);
	}

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 9) && (interface == 2))
		return usbua100_assign_midi_device(device, "UM-1", 0x81, 0x01, 0x00, 64);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 0))
		printk(KERN_INFO "USBUA100: Unsupported UA-100 interface %d\n", interface);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 3))
		printk(KERN_INFO "USBUA100: Unsupported SC-8850 interface %d\n", interface);

	else if ((device->descriptor.idVendor == 0x582) && (device->descriptor.idProduct == 9))
		printk(KERN_INFO "USBUA100: Unsupported UM-1 interface %d\n", interface);

	return NULL;
}

static void usbua100_disconnect(struct usb_device *device, void *driver_context)
{
	struct usbua100_midi_device *midi_device = (struct usbua100_midi_device*) driver_context;
	struct usbua100_audio_device *audio_device = (struct usbua100_audio_device*) driver_context;

	if ((audio_device >= usbua100_audio_devices) && (audio_device < &usbua100_audio_devices[USBUA100_MAX_AUDIO_DEVICES]))
		usbua100_release_audio_device(audio_device);

	else if ((midi_device >= usbua100_midi_devices) && (midi_device < &usbua100_midi_devices[USBUA100_MAX_AUDIO_DEVICES]))
	{
		if (strncmp(midi_device->device_name, "UA-100", 6) == 0)
		{
			usbua100_release_midi_device(midi_device);
			usbua100_release_midi_device(midi_device - 1);
			usbua100_release_midi_device(midi_device - 2);
		}
		else if (strncmp(midi_device->device_name, "SC-8850", 7) == 0)
		{
			usbua100_release_midi_device(midi_device);
			usbua100_release_midi_device(midi_device - 1);
			usbua100_release_midi_device(midi_device - 2);
			usbua100_release_midi_device(midi_device - 3);
		}

		else
			usbua100_release_midi_device(midi_device);
	}
}

static struct usb_driver usbua100_driver_registration =
{
	name: "usbua100",
	probe: usbua100_probe,
	disconnect: usbua100_disconnect,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	id_table:	NULL,
	driver_list: {NULL, NULL},
	fops: NULL
#endif
};


/************************* Module Entry and Exit *****************************/

int init_module(void)
{
	int status = 0;

	printk(KERN_INFO "USBUA100: Linux USB Driver for Roland MIDI interfaces v2001.06.30\n");

	memset(&usbua100_midi_devices, 0, sizeof(usbua100_midi_devices));
	memset(&usbua100_audio_devices, 0, sizeof(usbua100_midi_devices));

	if ((status = usb_register(&usbua100_driver_registration)) < 0)
		printk(KERN_ERR "USBUA100: Driver failed usb_register with code %d\n", status);

	return 0;
}

int cleanup_module(void)
{
	usb_deregister(&usbua100_driver_registration);

	printk(KERN_INFO "USBUA100: Module deregistered and unloaded\n");

	return 0;
}



